# Modified by Dmytro Kalpakchi, 2020
# Adapted from spaCy's built in visualization suite for dependencies and named entities.
# https://github.com/explosion/spaCy/tree/master/spacy/displacy

def is_in_jupyter():
    """Check if user is running spaCy from a Jupyter notebook by detecting the
    IPython kernel. Mainly used for the displaCy visualizer.
    RETURNS (bool): True if in Jupyter, False if not.
    """
    # https://stackoverflow.com/a/39662359/6400719
    try:
        shell = get_ipython().__class__.__name__
        if shell == "ZMQInteractiveShell":
            return True  # Jupyter notebook or qtconsole
    except NameError:
        return False  # Probably standard Python interpreter
    return False


def minify_html(html):
    """Perform a template-specific, rudimentary HTML minification for displaCy.
    Disclaimer: NOT a general-purpose solution, only removes indentation and
    newlines.
    html (unicode): Markup to minify.
    RETURNS (unicode): "Minified" HTML.
    """
    return html.strip().replace("    ", "").replace("\n", "")


def escape_html(text):
    """Replace <, >, &, " with their HTML encoded representation. Intended to
    prevent HTML errors in rendered displaCy markup.
    text (unicode): The original text.
    RETURNS (unicode): Equivalent text to be safely used within HTML.
    """
    text = text.replace("&", "&amp;")
    text = text.replace("<", "&lt;")
    text = text.replace(">", "&gt;")
    text = text.replace('"', "&quot;")
    return text



# coding: utf8

# Modified by Dmytro Kalpakchi, 2020
# Adapted from spaCy's built in visualization suite for dependencies and named entities.
# https://github.com/explosion/spaCy/tree/master/spacy/displacy

from __future__ import unicode_literals

import re

from udon2 import Node
import svgling

from .render import DependencyRenderer
from .utils import is_in_jupyter


_html = {}
RENDER_WRAPPER = None

def render_dep_tree(node, fname, img_format='svg', page=False, minify=False):
    if (img_format == 'svg'):
        with open(fname, 'w') as f:
            f.write(render(node, page=page, minify=minify))


def tree2list(node):
    if not node.children:
        return node.form
    lst = [node.form]
    for x in node.children:
        res = tree2list(x)
        if res:
            lst.append(res)
    return lst


def render_tree(node, fname, img_format='svg'):
    base_pixl = 16
    options = svgling.core.TreeOptions()
    with open(fname, 'w') as f:
        svg_string = svgling.core.TreeLayout(tree2list(node), options=options)._repr_svg_()
        ems = set(re.findall(r"\d+.?\d*em", svg_string))
        for em in ems:
            px = float(em[:-2]) * base_pixl
            svg_string = svg_string.replace(em, "{}px".format(px))
        f.write(svg_string)


def render(
    node, page=False, minify=False, jupyter=None, options={}, manual=False
):
    """Render displaCy visualisation.

    docs (list or Doc): Document(s) to visualise.
    style (unicode): Visualisation style, 'dep' or 'ent'.
    page (bool): Render markup as full HTML page.
    minify (bool): Minify HTML markup.
    jupyter (bool): Override Jupyter auto-detection.
    options (dict): Visualiser-specific options, e.g. colors.
    manual (bool): Don't parse `Doc` and instead expect a dict/list of dicts.
    RETURNS (unicode): Rendered HTML markup.

    DOCS: https://spacy.io/api/top-level#displacy.render
    USAGE: https://spacy.io/usage/visualizers
    """
    if isinstance(node, Node):
        docs = [node]
    if not all(isinstance(obj, Node) for obj in docs):
        raise ValueError("Invalid object passed to displaymaster: can work only with udon2.Node")
    renderer = DependencyRenderer(options=options)
    parsed = [parse_deps(doc, options) for doc in docs] if not manual else docs
    _html["parsed"] = renderer.render(parsed, page=page, minify=minify).strip()
    html = _html["parsed"]
    if RENDER_WRAPPER is not None:
        html = RENDER_WRAPPER(html)
    if jupyter or (jupyter is None and is_in_jupyter()):
        # return HTML rendered by IPython display()
        # See #4840 for details on span wrapper to disable mathjax
        from IPython.core.display import display, HTML

        return display(HTML('<span class="tex2jax_ignore">{}</span>'.format(html)))
    return html


def parse_deps(orig_doc, options={}):
    """Generate dependency parse in {'words': [], 'arcs': []} format.

    doc (Doc): Document do parse.
    RETURNS (dict): Generated dependency parse keyed by words and arcs.
    """
    doc = orig_doc # might reparse it later
    words = [{"text": w.form, "tag": w.upos, "morph": str(w.feats).split("|")} for w in doc.linear()]
    id2index = {w.id : i for i, w in enumerate(doc.linear())}
    arcs = []
    for word in doc.linear():
        if word.parent:
            i, governor = id2index[word.id], id2index[word.parent.id]
            if i < governor:
                arcs.append({
                    "start": i,
                    "end": governor,
                    "label": word.deprel,
                    "dir": "left"
                })
            elif i > governor:
                arcs.append({
                    "start": governor,
                    "end": i,
                    "label": word.deprel,
                    "dir": "right",
                })
    return {"words": words, "arcs": arcs, "settings": get_doc_settings(orig_doc)}


def set_render_wrapper(func):
    """Set an optional wrapper function that is called around the generated
    HTML markup on displacy.render. This can be used to allow integration into
    other platforms, similar to Jupyter Notebooks that require functions to be
    called around the HTML. It can also be used to implement custom callbacks
    on render, or to embed the visualization in a custom page.

    func (callable): Function to call around markup before rendering it. Needs
        to take one argument, the HTML markup, and should return the desired
        output of displacy.render.
    """
    global RENDER_WRAPPER
    if not hasattr(func, "__call__"):
        raise ValueError(Errors.E110.format(obj=type(func)))
    RENDER_WRAPPER = func


def get_doc_settings(doc):
    return {
        "direction": "ltr" # just set this as the only available option
    }



# coding: utf8
# Adapted from spaCy's built in visualization suite for dependencies and named entities.
# https://github.com/explosion/spaCy/tree/master/spacy/displacy

from __future__ import unicode_literals


# Setting explicit height and max-width: none on the SVG is required for
# Jupyter to render it properly in a cell

TPL_DEP_SVG = """
<svg xmlns="http://www.w3.org/2000/svg" xmlns:xlink="http://www.w3.org/1999/xlink" id="{id}" width="{width}" height="{height}" direction="{dir}" style="max-width: none; height: {height}px; color: {color}; background: {bg}; font-family: {font}; direction: {dir}">{content}</svg>
"""


TPL_DEP_WORDS = """
<text class="token" fill="currentColor" text-anchor="middle" y="{y}">
    <tspan class="word" fill="currentColor" x="{x}" font-size="2em">{text}</tspan>
    <tspan class="tag" dy="2em" fill="currentColor" x="{x}" font-size="2em">{tag}</tspan>
    {morph}
</text>
"""

TPL_DEP_MORPH = """
<tspan class="tag" dy="{dy}" fill="currentColor" x="{x}" font-size="1.75em">{morph}</tspan>
"""


TPL_DEP_ARCS = """
<g class="arrow">
    <path class="arc" id="arrow-{id}-{i}" stroke-width="{stroke}px" d="{arc}" fill="none" stroke="currentColor"/>
    <text dy="1.25em" style="font-size: 1.75em; letter-spacing: 1px">
        <textPath xlink:href="#arrow-{id}-{i}" class="label" startOffset="50%" side="{label_side}" fill="currentColor" text-anchor="middle">{label}</textPath>
    </text>
    <path class="arrowhead" d="{head}" fill="currentColor"/>
</g>
"""


TPL_FIGURE = """
<figure style="margin-bottom: 6rem">{content}</figure>
"""

TPL_TITLE = """
<h2 style="margin: 0">{title}</h2>
"""

TPL_PAGE = """
<!DOCTYPE html>
<html>
    <head>
        <title>displaCy</title>
    </head>

    <body style="font-size: 16px; font-family: -apple-system, BlinkMacSystemFont, 'Segoe UI', Helvetica, Arial, sans-serif, 'Apple Color Emoji', 'Segoe UI Emoji', 'Segoe UI Symbol'; padding: 4rem 2rem; direction: {dir}">{content}</body>
</html>
"""



import os
import sys
import codecs
import glob
import platform
import setuptools

from sysconfig import get_paths


BOOST_DIR = os.environ['BOOST_DIR']
IS_WINDOWS = platform.system().lower() == 'windows'
IS_LINUX = platform.system().lower() == 'linux'
C_SRC = 'udon2_cpp_src'
README = "README.md"


def read(rel_path):
    here = os.path.abspath(os.path.dirname(__file__))
    with codecs.open(os.path.join(here, rel_path), 'r') as fp:
        return fp.read()


def get_version(rel_path):
    for line in read(rel_path).splitlines():
        if line.startswith('__version__'):
            delim = '"' if '"' in line else "'"
            return line.split(delim)[1]
    else:
        raise RuntimeError("Unable to find version string.")


vinfo = sys.version_info
path_info = get_paths()


with open(README, "r") as fh:
    long_description = fh.read()

if IS_WINDOWS:
    arch = platform.architecture()[0][:2]
    boost_library = f'boost_python{vinfo.major}{vinfo.minor}*'
    boost_include = [BOOST_DIR]
    boost_lib = glob.glob(os.path.join(BOOST_DIR, f'lib{arch}-msvc-*'))
elif IS_LINUX:
    arch = platform.machine()
    syst = platform.system().lower()
    boost_library = f'boost_python{vinfo.major}{vinfo.minor}'
    boost_include = [os.path.join(BOOST_DIR, 'include')]
    boost_lib = [os.path.join(BOOST_DIR, 'lib'), os.path.join(BOOST_DIR, 'lib', f"{arch}-{syst}-gnu")]
else:
    raise RuntimeError("The platform is not supported!")

include_dirs = [path_info['include']] + boost_include
library_dirs = [path_info['stdlib']] + boost_lib
libraries = [boost_library]

core_module = setuptools.Extension(
    'udon2.core',
    sources=[
        os.path.join(C_SRC, 'Util.cpp'),
        os.path.join(C_SRC, 'Node.cpp'),
        os.path.join(C_SRC, 'MultiWordNode.cpp'),
        os.path.join(C_SRC, 'ConllReader.cpp'),
        os.path.join(C_SRC, 'ConllWriter.cpp'),
        os.path.join(C_SRC, 'udon2_core.cpp')
    ],
    include_dirs=include_dirs,
    library_dirs=library_dirs,
    libraries=libraries
)

def compiled_shared_obj_files_iter(fnames):
    compiled_obj = []
    if IS_WINDOWS:
        for f in fnames:
            compiled_obj += glob.glob(os.path.join('build', f'temp*-{vinfo.major}.{vinfo.minor}', 'Release', C_SRC, '{}.obj'.format(f)))
    else:
        for f in fnames:
            compiled_obj += glob.glob(os.path.join('build', f'temp*-{vinfo.major}.{vinfo.minor}', C_SRC, '{}.o'.format(f)))
    for x in compiled_obj:
        yield x

transformations_module = setuptools.Extension(
    'udon2.transform',
    sources=[
        os.path.join(C_SRC, 'transformations.cpp'),
        os.path.join(C_SRC, 'udon2_transform.cpp')
    ],
    extra_objects=compiled_shared_obj_files_iter(['Util', 'MultiWordNode', 'Node']), # sort of lazy eval here to reuse the already compiled files
    include_dirs=include_dirs,
    library_dirs=library_dirs,
    libraries=libraries
)

kernels_module = setuptools.Extension(
    'udon2.kernels',
    sources=[
        os.path.join(C_SRC, 'kernels.cpp'),
        os.path.join(C_SRC, 'udon2_kernels.cpp')
    ],
    extra_objects=compiled_shared_obj_files_iter(['Util', 'MultiWordNode', 'Node', 'transformations']), # sort of lazy eval here to reuse the already compiled files
    include_dirs=include_dirs,
    library_dirs=library_dirs,
    libraries=libraries
)

setuptools.setup(
    name="udon2", # Replace with your own username
    version=get_version("src/udon2/__init__.py"),
    author="Dmytro Kalpakchi",
    author_email="dmytroka@kth.se",
    description="Prepare your UD trees to be served!",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/dkalpakchi/udon2",
    packages=setuptools.find_packages('src'),
    package_dir={'': 'src'},
    classifiers=[
        "Intended Audience :: Science/Research",
        "Programming Language :: Python :: 3 :: Only",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "License :: OSI Approved :: MIT License",
        "Operating System :: POSIX :: Linux",
        "Operating System :: Microsoft :: Windows",
        "Topic :: Scientific/Engineering"
    ],
    ext_modules=[core_module, transformations_module, kernels_module],
    python_requires='>=3.6',
    install_requires=["six", "svgling"]
)



# An attempt to emulate auditwheel for Windows
# Author: Dmytro Kalpakchi, 2020
import os
import glob
import sys
import platform
import argparse
import shutil
import sysconfig
import zipfile
import hashlib
from pathlib import Path

import pefile
from machomachomangler.pe import redll

if __name__ == '__main__':
    parser = argparse.ArgumentParser()
    parser.add_argument('-p', '--package', type=str, help='The name of your package (to be used in name mangling)')
    parser.add_argument('-v', '--version', type=str, help='The version of your package (to be used in name mangling)')
    args = parser.parse_args()

    if platform.system().lower() != 'windows':
        print('Error: This tool only supports Windows')
        sys.exit(1)

    path_info = sysconfig.get_paths()
    arch = os.environ["PYTHON_ARCH"]

    print("winauditwheel started...")

    # 1. Find DLL names
    dlls = set()
    pyd_files = glob.glob(os.path.join('build', 'lib*', args.package, '*.pyd'))
    print("Checking .pyd files:", pyd_files)
    for dll_fname in pyd_files:
        p = pefile.PE(dll_fname)
        for entry in p.DIRECTORY_ENTRY_IMPORT:
            dlls.add(entry.dll)
    print("Found DLL names", dlls)

    # 2. Find the actual DLL files
    search_paths = list([os.environ['BOOST_DIR'], path_info['stdlib']] + sys.path)
    redistribute = {}
    for dll in dlls:
        sdll = dll.decode('ascii')
        if 'KERNEL' in sdll or 'api-ms' in sdll or sdll.startswith('python'): continue

        for d in search_paths:
            for f in glob.glob(os.path.join(d, '**', sdll), recursive=True):
                redistribute[dll] = f

        for d in os.environ['PATH'].split(';'):
            for f in glob.glob(os.path.join(d, sdll)):
                p = pefile.PE(f)
                machine_type = pefile.MACHINE_TYPE[p.FILE_HEADER.Machine]
                if arch == '64' and machine_type != 'IMAGE_FILE_MACHINE_AMD64': continue
                if arch == '32' and machine_type != 'IMAGE_FILE_MACHINE_I386': continue
                redistribute[dll] = f
    print("Found actual DLLs", redistribute)

    # 3. Copy the DLLs to the current folder and change names
    salt = hashlib.sha1(str.encode("{}_{}".format(args.package, args.version))).hexdigest()
    dll_map = {}
    for bfname, src in redistribute.items():
        fname = bfname.decode('ascii')
        fn, ext = os.path.splitext(fname)
        dest = "{}.{}{}".format(fn, salt, ext)
        shutil.copyfile(src, os.path.join('dist', dest))
        dll_map[bfname] = dest.encode("ascii")
    print("DLL name mangling map", dll_map)

    # 4. Repair a wheel
    DIST_DIR = 'dist'
    for whl in glob.glob(os.path.join(DIST_DIR, '*.whl')):
        whl_repair_dir = os.path.join(DIST_DIR, 'whl_repair')
        with zipfile.ZipFile(whl, 'r') as zip_ref:
            zip_ref.extractall(whl_repair_dir)

        for pyd_file in glob.glob(os.path.join(whl_repair_dir, '**', '*.pyd'), recursive=True):
            fn, ext = os.path.splitext(pyd_file)

            with open(pyd_file, "rb") as f:
                buf = f.read()

            new_buf = redll(buf, dll_map)

            with open(pyd_file, "wb") as f:
                f.write(new_buf)

        with zipfile.ZipFile(whl, 'w') as new_whl:
            for folder, subfolders, filenames in os.walk(whl_repair_dir):
                for filename in filenames:
                    #create complete filepath of file in directory
                    file_path = os.path.join(folder, filename)
                    # Add file to zip
                    new_whl.write(file_path, os.path.join(folder.replace(whl_repair_dir, ''), filename))
            whl_path = Path(whl)
            for dll_file in glob.glob(os.path.join(whl_path.parent, '*.dll')):
                new_whl.write(dll_file, os.path.join(args.package, os.path.basename(dll_file)))

        # Clean up everything except a new wheel
        empty_folders = []
        for folder, subfolders, filenames in os.walk(DIST_DIR):
            for filename in filenames:
                if os.path.splitext(filename)[1] != '.whl':
                    file_path = os.path.join(folder, filename)
                    os.remove(file_path)
            if folder != DIST_DIR:
                empty_folders.append(folder)

        empty_folders.reverse()
        for fd in empty_folders:
            os.rmdir(fd)

    print("winauditwheel finished...")
